//===========================================================================
//
//
//  CRYSTALFONTZ CFAH1602B 16x2 Character Display
//
//  This code uses the 8-bit parallel MCU mode of the display.
//
//  https://www.crystalfontz.com/product/cfah1602b family
//
//  The controller is a Sitronix ST7066U:
//    https://www.crystalfontz.com/controllers/Sitronix/ST7066U
//
//  Seeeduino v4.2, an open-source 3.3v capable Arduino clone.
//    https://www.seeedstudio.com/Seeeduino-V4.2-p-2517.html
//    https://github.com/SeeedDocument/SeeeduinoV4/raw/master/resources/Seeeduino_v4.2_sch.pdf
//============================================================================
//
//
//
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//============================================================================



//============================================================================
// LCD & USD control lines
//   ARD      | Port | CFA1602Y   |  Function                   | Wire
//------------+------+------------+-----------------------------+------------
//  5V        |      | #2         |  POWER 5V                   | Red
//  GND       |      | #1         |  GROUND                     | Black
//  3.3V      |      | #15        |  POWER 3.3V                 | Green
//  GND       |      | #16        |  GROUND                     | Black
//  N/A       |      | #3         |  POWER ~1V                  | Orange
// -----------+------+------------+-----------------------------+------------
//  #8/D8     |  PB0 | #4         |  Data/Instruction   (RS)    | Purple
//  #9/D9     |  PB1 | #6         |  Chip Select/Enable (CS/E)  | Yellow
// #10/D10    |  PB2 | #5         |  Read/Write         (RW)    | Grey
// -----------+------+------------+-----------------------------+------------
//  #0/D0     |  PD0 | #7         |  LCD_D10 (DB0)              | Black
//  #1/D1     |  PD1 | #8         |  LCD_D11 (DB1)              | Brown
//  #2/D2     |  PD2 | #9         |  LCD_D12 (DB2)              | Red
//  #3/D3     |  PD3 | #10        |  LCD_D13 (DB3)              | Orange
//  #4/D4     |  PD4 | #11        |  LCD_D14 (DB4)              | Yellow
//  #5/D5     |  PD5 | #12        |  LCD_D15 (DB5)              | Green
//  #6/D6     |  PD6 | #13        |  LCD_D16 (DB6)              | Blue
//  #7/D7     |  PD7 | #14        |  LCD_D17 (DB7)              | Purple
// -----------+------+------------+-----------------------------+------------
//============================================================================

#define CLR_RS     (PORTB &= ~(0x01)) //pin #8  - Data/Instruction
#define SET_RS     (PORTB |=  (0x01)) //pin #8  - Data/Instruction
#define CLR_E      (PORTB &= ~(0x02)) //pin #9  - Chip Enable Signal
#define SET_E      (PORTB |=  (0x02)) //pin #9  - Chip Enable Signal
#define CLR_RW     (PORTB &= ~(0x04)) //pin #10 - Read/Write
#define SET_RW     (PORTB |=  (0x04)) //pin #10 - Read/Write

#define LCD_DATA   (PORTD) //PORTD is used for the 8-bit parallel interface

#include <Arduino.h>

unsigned char CGRAM1[8] ={0x04,0x0E,0x15,0x04,0x04,0x04,0x04,0x04,};  // ?`
unsigned char CGRAM2[8] ={0x04,0x04,0x04,0x04,0x04,0x15,0x0e,0x04,};  // ?`
unsigned char CGRAM3[8] ={0x1f,0x11,0x11,0x11,0x11,0x11,0x11,0x1f,};  // ?`
unsigned char CGRAM4[8] ={0x0a,0x15,0x0a,0x15,0x0a,0x15,0x0a,0x15,};  // ?`

//============================================================================
// writeCommand(uint8_t command);
// writeCommandNB(uint8_t command);
// uint8_t writeData(uint8_t data);
// uint8_t checkBusy();
// void init_1602b();
// void init_st7066u();
// void writeString(char* myString);
// void CGRAM();

//============================================================================
void writeCommand(uint8_t command)
{
	checkBusy();

	//Put the command on the port
	LCD_DATA = command;
	
	// Select the LCD's command register
	CLR_RS;

	CLR_RW;

	// Deselect the LCD controller
	SET_E;
	CLR_E;

}
//============================================================================
// void writeCommandNB(uint8_t command)
// {
// 	//checkBusy();

// 	//Put the command on the port
// 	LCD_DATA = command;
	
// 	// Select the LCD's command register
// 	CLR_RS;

// 	CLR_RW;

// 	// Deselect the LCD controller
// 	SET_E;
// 	CLR_E;

// }

//----------------------------------------------------------------------------
uint8_t writeData(uint8_t data)
{
	checkBusy();
	//Put the data on the port
	LCD_DATA = data;

	// Select the LCD's command register
	SET_RS;

	CLR_RW;

	// Deselect the LCD controller
	SET_E;
	CLR_E;
	
}

uint8_t checkBusy()
{
  //Turn Busy Flag on and turn PORTD to INPUT
	PORTD = 0x80;
	DDRD = 0x00;
	
	CLR_RS;
	SET_RW;

	// Select the LCD controller
	SET_E;

	//Watch Pin 7 for the busy flag
	while (0x80 == (PIND & 0x80))
	{
		delayMicroseconds(80);
		CLR_E;
		delayMicroseconds(80);
		SET_E;
	}
	CLR_E;
	CLR_RW;

  //turn PORTD to OUTPUT
	DDRD = 0xFF;

	return LCD_DATA;
}

// void init_1602b()
// {

// 	//The following block is the power on sequence. Each time the module is power
// 	//cycled, this is the code that is initiated internally.
// 	//This is for reference only, it is not necessary unless you are changing modes
// 	//without power cycling the controller

// 	//------------------------------------------------------------------------
// 	//Start Power On Sequence
// 	//Display clear
// 	writeCommand(0x01);
// 	//Function Set
// 	writeCommand(0x30);	
// 		//	00110000
// 		//	0011XXXX
// 		//	    ||||---	Font Table Selection
// 		// 	    ||		>>00: English_Japenese character font table
// 		//		||		  01: Western European character font table - I
// 		//		||		  10: English_Russian character font table
// 		//		||		  11: Western European character font table - II
// 		//		||-----	Character Font Set
// 		//		|		>>0: 5x8 dot character font
// 		//		|		  1: 5x10 dot character font
// 		//		|------	Interface Data Length Control Bit	
// 		//				>>0: data is sent or received in 4-bit length DB4-DB7
// 		//				  1: data is sent or received in 8-bit length DB4-DB7


// 	delayMicroseconds(37);
// 	//Cursor / Display Shift Instruction (2 parts)
// 	//only sending second part
// 	writeCommand(0x13);	
// 		//	00010011
// 		//	0001XX11
// 		//		||----	Enable / Disable Internal Power
// 		//		|		>>0: Internal Power OFF
// 		//		|	 	  1: Internal Power ON
// 		//		|-----	Graphic Mode / Character Mode Selection
// 		//				>>0: Character Mode
// 		//				  1: Graphic Mode

// 	delayMicroseconds(37);
// 	//Display ON/OFF Control
// 	writeCommand(0x08);	
// 		//	00001000	
// 		//	00001XXX
// 		//		 |||--	Blinking Control Bit
// 		//		 ||		  1: Cursor Blinking ON
// 		//		 ||		>>0: Cursor Blinking OFF
// 		//		 ||---	Cursor Display Control
// 		//		 |		>>0: Cursor OFF
// 		//		 |		  1: Cursor ON
// 		//		 |----	Display ON/OFF
// 		//		  		>>0: Display OFF
// 		//		  		  1: Display ON

// 	delayMicroseconds(37);
// 	//Entry Mode Set
// 	writeCommand(0x04); 
// 		//	00000100
// 	   //	000001XX
// 	   //		  ||--	Shift Entire Display Control Bit
// 	   //		  |		>>0: Decrement DDRAM Address by 1 when a character
// 	   //		  |			 code is written into or read from DDRAM
// 	   //		  |		  1: Increment DDRAM Address by 1 when a character
// 	   //		  |			 code is written into or read from DDRAM
// 	   //		  |---	Increment/Decrement bit
// 	   //		  		>>0: when writing to DDRAM, each
// 	   //		  			 entry moves the cursor to the left
// 	   //			 	  1: when writing to DDRAM, each 
// 	   //		  			 entry moves the cursor to the right

//    //End Power On Sequence
//    //------------------------------------------------------------------------



// 	//Cursor / Display Shift Instruction (2 parts)
// 	writeCommand(0x14);
// 		//	00010100
// 		//	0001XX00
// 		//		||----	Shift Function
// 		//				  00: Shifts cursor position to the left
// 		//					  (AC is decremented by 1)
// 		//				>>01: Shifts cuursor position to the right
// 		//					  (AC is incremented by 1)
// 		//				  10: Shifts entire display to the left
// 		//					  The cursor follows the display shift
// 		//				  11: Shifts entire display to the right
// 		//					  The cursor follows the display shift
// 	writeCommand(0x17);
// 		//	00010111
// 		//	0001XX11
// 		//		||----	Enable / Disable Internal Power
// 		//		|		  0: Internal Power OFF
// 		//		|	 	>>1: Internal Power ON
// 		//		|-----	Graphic Mode / Character Mode Selection
// 		//				>>0: Character Mode
// 		//				  1: Graphic Mode
// 	delay(500);

// 	checkBusy();
// 	//Function Set
// 	writeCommand(0x3B);	//00111000
// 		//	001XXX00
// 		//	   |||-----	Character Font Set
// 		//	   ||		>>0: 5x8 dot character font
// 		//	   ||		  1: 5x10 dot character font
// 		//	   ||------	Display line number control bit	
// 		//	   |		  0: 1-line display mode is set
// 		//	   |		>>1: 2-line display mode is set
// 		//	   |-------	Interface data length control bit
// 		//				  0: 4-bit bus mode with MPU
// 		//				>>1: 8-bit bus mode with MPU

// 	//Display ON/OFF Control
// 	writeCommand(0x0C);	//0001100
// 		//	00001XXX
// 		//		 |||--	Blinking Control Bit
// 		//		 ||		  1: Cursor Blinking ON
// 		//		 ||		>>0: Cursor Blinking OFF
// 		//		 ||---	Cursor Display Control
// 		//		 |		>>0: Cursor OFF
// 		//		 |		  1: Cursor ON
// 		//		 |----	Display ON/OFF
// 		//		  		  0: Display OFF
// 		//		  		>>1: Display ON

// 	//Display Clear 00000001
// 	writeCommand(0x01);

// 	//Return Home 00000010
// 	writeCommand(0x02);

// 	writeCommand(0x17);
// 	//Graphic vs character mode setting, RS=0,R/W=0
// 	// 7654 3210
// 	// 0001 GP11
// 	//  G = Mode: 1=graphic, 0=character
// 	//  C = Power: 1=0n, 0=off

// 	//Entry Mode Set
// 	writeCommand(0x06); //00000110
//       //  000001XX
//       //        ||--  Shift Entire Display Control Bit
//       //        |     >>0: Decrement DDRAM Address by 1 when a character
//       //        |          code is written into or read from DDRAM
//       //        |       1: Increment DDRAM Address by 1 when a character
//       //        |          code is written into or read from DDRAM
//       //        |
//       //        |---	Increment/Decrement bit
//       //                0: when writing to DDRAM, each
//       //                   entry moves the cursor to the left
//       //              >>1: when writing to DDRAM, each 
//       //                   entry moves the cursor to the right

// 	//set DDRAM
// 	writeCommand(0x80);


// }

void init_st7066u()
{
	// Function SET
	//	DL:interface data is 8/4 bits
	//		DL = 0; 4-Bit interface
	//		DL = 1; 8-Bit interface
	//	N:number of line is 2/1
	//		N = 0 ; 1-line display
	//		N = 1 ; 2-line display
	//	F:font size is 5x11/5x8
	//		F = 0; 5x8 dot character font
	//		F = 1; 5x11 dot character font
	//| RS | RW |	DB7 | DB6 | DB5 | DB4 | DB3 | DB2 | DB1 | DB0 |
	//|  0 |  0 |	 0  |  0  |  1  |  DL |  N  |  F  |  x  |  x  |
	//|  0 |  0 |  0  |  0  |  1  |  1  |  1  |  0  |  0  |  0  | = 0b0011 1000 = 0x38
	writeCommand(0x38);	

	// Send it again?
	writeCommand(0x38);
	writeCommand(0x38);
	writeCommand(0x38);

	// Display ON/OFF Control
	//	D = 0/1; Display off/on
	//	C = 0/1; Cursor off/on
	//	B = 0/1; Cursor Blinking off/on
	//| RS | RW |	DB7 | DB6 | DB5 | DB4 | DB3 | DB2 | DB1 | DB0 |
	//|  0 |  0 |	 0  |  0  |  0  |  0  |  1  |  D  |  B  |  C  |
	//|  0 |  0 |	 0  |  0  |  0  |  0  |  1  |  0  |  0  |  0  | = 0b00001000 = 0x08
	writeCommand(0x08);
	//|  0 |  0 |	 0  |  0  |  0  |  0  |  1  |  1  |  1  |  1  | = 0b00001111 = 0x0F
	writeCommand(0x0F);

	//Display Clear
	//| RS | RW |	DB7 | DB6 | DB5 | DB4 | DB3 | DB2 | DB1 | DB0 |
	//|  0 |  0 |	 0  |  0  |  0  |  0  |  0  |  0  |  0  |  1  | = 0b00000000 = 0x01
	writeCommand(0x01);

	//Entry Mode Set
	//	I/D = 0; Decrement by 1
	//	I/D = 1; Increment by 1
	//	S = 0; No shift
	//	S = 1; shift left or right depending on I/D
	//| RS | RW |	DB7 | DB6 | DB5 | DB4 | DB3 | DB2 | DB1 | DB0 |
	//|  0 |  0 |	 0  |  0  |  0  |  0  |  0  |  1  | I/D |  S  | = 0b00000110 = 0x06
	writeCommand(0x06);
	
}
void writeString(char* myString)
{
	uint8_t i = 0;
	do
	{
    //send one character at a time
		writeData((uint8_t)myString[i]);
		i++;
	} while (myString[i] != NULL);
}

void setup()
{
  //Setup ports - PORTD to OUTPUT
	DDRD = 0xFF;
	PORTD = 0x00;

  //Setup ports - PORTB - PINB 0,1,2,4 to OUTPUT
	DDRB = 0x07;
	PORTB = 0x00;
	delay(500);

	CLR_RW;
//	init_1602b();
	init_st7066u();
	delay(10);
}

void CGRAM()
{
	unsigned char i,j;
	writeCommand(0x40);    //??????CGRAM Address?????000000(0x00)
					   //CGRAM?????0x40+0x00=0x40
	for(i = 0;i<8;i++)
	{
	   	writeData(CGRAM3[i]);
	}
	writeCommand(0x48);	   //??????CGRAM Address?????001000(0x08)
					   //CGRAM?????0x40+0x08=0x48
	for(j = 0;j<8;j++)
	{
	   	writeData(CGRAM4[j]); //data write to CGRAM
	}
}

void loop()
{
	//Call the CGRAM setup
	CGRAM();

	delay(500);

	//clear screen
//	writeCommand(0x01);

	//set DDRAM address to first row first column
	writeCommand(0x80);
	writeString("* CrystalFontz *");
	//set DDRAM address to second row first column
	writeCommand(0xC0);
	writeString("* CFAL1602BGGH *");

	delay(1000);

  //set DDRAM address to first row first column
	writeCommand(0x80);
	writeString("0123456789ABCDEF");
  //set DDRAM address to second row first column
	writeCommand(0xC0);
	writeString("GHIJKLMNOPQRSTUV");

	delay(1000);
}
